import math
from compas.geometry import Point
from compas.geometry import Vector
from compas.geometry import Frame
from compas.geometry import Rotation
from compas.geometry import Sphere
from compas.geometry import intersection_sphere_sphere
from compas.geometry import intersection_plane_circle
from compas.geometry import tangent_points_to_circle_xy

# TODO: This is very slow...
# can this be improved? for example https://github.com/visose/Robots/blob/master/Robots/Kinematics.cs


def forward_kinematics_spherical_wrist(joint_values, points):
    """Forward kinematics function for spherical wrist robots.

    Parameters
    ----------
    joint_values : list of float
        A list of 6 joint values in radians.
    points : list of point
        A list of 4 points specifying the robot's joint positions.

    Returns
    -------
    :class:`compas.geometry.Frame`

    Notes
    -----
    This is a modification of the *Lobster* tool for solving the inverse
    kinematics problem for a spherical wrist robots.

    https://www.grasshopper3d.com/forum/topics/lobster-reloaded?groupUrl=lobster&

    This is the instruction on how to determine the 4 points.
    http://archive.fabacademy.org/archives/2017/fablabcept/students/184/assets/lobster_ik.pdf

    Check that p2, p3, and p4 are all in the XZ plane (y coordinates are zero)!
    """

    p1, p2, p3, p4 = points

    a1, a2, a3, a4, a5, a6 = joint_values

    af1 = Frame.worldZX()
    af1.point = p1
    af2 = Frame.worldZX()
    af2.point = p2
    af3 = Frame.worldYZ()
    af3.point = p3
    af4 = Frame.worldZX()
    af4.point = p3
    af5 = Frame.worldYZ()
    af5.point = p3
    af6 = Frame.worldYZ()
    af6.point = p4

    axis_frames = [af1, af2, af3, af4, af5, af6]

    # TODO: this is slow
    Rot1 = Rotation.from_axis_and_angle((0, 0, 1), -1 * a1)
    for i in range(0, 6):
        axis_frames[i].transform(Rot1)

    Rot2 = Rotation.from_axis_and_angle(axis_frames[0].zaxis, a2 + math.pi / 2, point=axis_frames[0].point)
    for i in range(1, 6):
        axis_frames[i].transform(Rot2)

    Rot3 = Rotation.from_axis_and_angle(axis_frames[1].zaxis, a3 - math.pi / 2, point=axis_frames[1].point)
    for i in range(2, 6):
        axis_frames[i].transform(Rot3)

    Rot4 = Rotation.from_axis_and_angle(axis_frames[2].zaxis, a4 * -1.0, axis_frames[2].point)
    for i in range(3, 6):
        axis_frames[i].transform(Rot4)

    Rot5 = Rotation.from_axis_and_angle(axis_frames[3].zaxis, a5, axis_frames[3].point)
    for i in range(4, 6):
        axis_frames[i].transform(Rot5)

    Rot6 = Rotation.from_axis_and_angle(axis_frames[4].zaxis, -1.0 * a6, axis_frames[4].point)
    for i in range(5, 6):
        axis_frames[i].transform(Rot6)

    return axis_frames[5]


def inverse_kinematics_spherical_wrist(target_frame, points):
    """Inverse kinematics function for spherical wrist robots.

    Parameters
    ----------
    frame : :class:`compas.geometry.Frame`
        The frame we search the inverse kinematics for.
    points : list of point
        A list of 4 points specifying the robot's joint positions.

    Returns
    -------
    list of list of float
        The 8 analytical IK solutions.


    Notes
    -----
    This is a modification of the *Lobster* tool for solving the inverse
    kinematics problem for a spherical wrist robots.

    https://www.grasshopper3d.com/forum/topics/lobster-reloaded?groupUrl=lobster&

    This is the instruction on how to determine the 4 points.
    http://archive.fabacademy.org/archives/2017/fablabcept/students/184/assets/lobster_ik.pdf

    Check that p2, p3, and p4 are all in the XZ plane (y coordinates are zero)!
    """
    p1, p2, p3, p4 = points

    end_frame = Frame(target_frame.point, target_frame.yaxis, target_frame.xaxis)

    wrist_offset = p4.x - p3.x
    lower_arm_length = p2.z - p1.z
    upper_arm_length = (p3 - p2).length
    pln_offset = math.fabs(p2.y - p1.y)
    axis4_offset_angle = math.atan2(p3.z - p2.z, p3.x - p2.x)
    wrist = end_frame.to_world_coordinates(Point(0, 0, wrist_offset))

    p1_proj = p1.copy()
    p1_proj.y = p2.y

    axis1_angles = []
    axis2_angles = []
    axis3_angles = []
    axis4_angles = []
    axis5_angles = []
    axis6_angles = []

    if pln_offset == 0:
        axis1_angle = -1 * math.atan2(wrist.y, wrist.x)
        if axis1_angle > math.pi:
            axis1_angle -= 2 * math.pi
        axis1_angles += [axis1_angle] * 4
        axis1_angle += math.pi
        if axis1_angle > math.pi:
            axis1_angle -= 2 * math.pi
        axis1_angles += [axis1_angle] * 4
    else:
        circle = (0, 0, 0), pln_offset
        point = (wrist.x, wrist.y, 0)
        t1, t2 = tangent_points_to_circle_xy(circle, point)
        a1 = Vector(0, 1, 0).angle_signed(t1, (0, 0, -1))
        a2 = Vector(0, 1, 0).angle_signed(t2, (0, 0, -1))
        axis1_angles += [a1] * 4
        axis1_angles += [a2] * 4

    for i in range(2):

        axis1_angle = axis1_angles[i * 4]
        Rot1 = Rotation.from_axis_and_angle([0, 0, 1], -1 * axis1_angle, point=[0, 0, 0])
        p1A = p1_proj.transformed(Rot1)
        elbow_dir = Vector(1, 0, 0).transformed(Rot1)
        sphere1 = Sphere(p1A, lower_arm_length)
        sphere2 = Sphere(wrist, upper_arm_length)
        elbow_frame = Frame(p1A, elbow_dir, [0, 0, 1])
        elbow_plane = (p1A, elbow_frame.normal)

        _, (center, radius, normal) = intersection_sphere_sphere(sphere1, sphere2)
        circle = ((center, normal), radius)

        intersect_pt1, intersect_pt2 = intersection_plane_circle(elbow_plane, circle)

        for j in range(2):
            if j == 0:
                elbow_pt = intersect_pt1
            else:
                elbow_pt = intersect_pt2
            elbow_pt = Point(*elbow_pt)

            elbowx, elbowy, _ = elbow_frame.to_local_coordinates(elbow_pt)
            wristx, wristy, _ = elbow_frame.to_local_coordinates(wrist)

            axis2_angle = math.atan2(elbowy, elbowx)
            axis3_angle = math.pi - axis2_angle + math.atan2(wristy - elbowy, wristx - elbowx) - axis4_offset_angle

            for k in range(2):
                axis2_angles.append(-axis2_angle)
                axis3_angle_wrapped = -axis3_angle + math.pi
                while (axis3_angle_wrapped >= math.pi):
                    axis3_angle_wrapped -= 2 * math.pi
                while (axis3_angle_wrapped < -math.pi):
                    axis3_angle_wrapped += 2 * math.pi
                axis3_angles.append(axis3_angle_wrapped)

            for k in range(2):
                axis4 = wrist - elbow_pt
                axis4.transform(Rotation.from_axis_and_angle(elbow_frame.zaxis, -axis4_offset_angle))
                temp_frame = elbow_frame.copy()
                temp_frame.transform(Rotation.from_axis_and_angle(temp_frame.zaxis, axis2_angle + axis3_angle))

                # // B = TempPlane
                axis4_frame = Frame(wrist, temp_frame.zaxis, temp_frame.yaxis * -1.0)
                axis6x, axis6y, axis6z = axis4_frame.to_local_coordinates(end_frame.point)

                axis4_angle = math.atan2(axis6y, axis6x)
                if k == 1:
                    axis4_angle += math.pi
                    if axis4_angle > math.pi:
                        axis4_angle -= 2 * math.pi

                axis4_angle_wrapped = axis4_angle + math.pi / 2
                while (axis4_angle_wrapped >= math.pi):
                    axis4_angle_wrapped -= 2 * math.pi
                while (axis4_angle_wrapped < -math.pi):
                    axis4_angle_wrapped += 2 * math.pi
                axis4_angles.append(axis4_angle_wrapped)

                axis5_frame = axis4_frame.copy()
                axis5_frame.transform(Rotation.from_axis_and_angle(axis4_frame.zaxis, axis4_angle))
                axis5_frame = Frame(wrist, axis5_frame.zaxis * -1, axis5_frame.xaxis)
                axis6x, axis6y, _ = axis5_frame.to_local_coordinates(end_frame.point)
                axis5_angle = math.atan2(axis6y, axis6x)
                axis5_angles.append(axis5_angle)

                axis6_frame = axis5_frame.copy()
                axis6_frame.transform(Rotation.from_axis_and_angle(axis5_frame.zaxis, axis5_angle))
                axis6_frame = Frame(wrist, axis6_frame.yaxis * -1, axis6_frame.zaxis)
                endx, endy, _ = axis6_frame.to_local_coordinates(end_frame.to_world_coordinates(Point(1, 0, 0)))
                axis6_angle = math.atan2(endy, endx)
                axis6_angles.append(axis6_angle)

    return [[a1, a2, a3, a4, a5, a6] for a1, a2, a3, a4, a5, a6 in zip(axis1_angles, axis2_angles, axis3_angles, axis4_angles, axis5_angles, axis6_angles)]



# -*- coding: utf-8 -*-

import math
from compas.geometry import Point
from .spherical_wrist import forward_kinematics_spherical_wrist
from .spherical_wrist import inverse_kinematics_spherical_wrist


class SphericalWristKinematics(object):
    """
    """

    def __init__(self, points):
        self.points = points

    def forward(self, joint_values):
        joint_values = self._pre_process(joint_values)
        return forward_kinematics_spherical_wrist(joint_values, self.points)

    def inverse(self, frame_rcf):
        solutions = inverse_kinematics_spherical_wrist(frame_rcf, self.points)
        return list(self._post_process(solutions))
        # return solutions

    def _post_process(self, solutions):
        return solutions

    def _pre_process(self, joint_values):
        return joint_values


class Staubli_TX260LKinematics(SphericalWristKinematics):
    """Analytical IK solver for the Stäubli TX2 60L robot."""
    def __init__(self):
        points = [Point(0.000, 0.000, 0.375),
                  Point(0.000, 0.020, 0.775),
                  Point(0.450, 0.020, 0.775),
                  Point(0.520, 0.020, 0.775)]
        super(Staubli_TX260LKinematics, self).__init__(points)

    def _pre_process(self, joint_values):
        q1, q2, q3, q4, q5, q6 = joint_values
        q1 = -1 * q1
        q2 = q2 - math.pi / 2
        q4 = q4 * -1
        q6 = q6 * -1 + math.pi / 2
        return [q1, q2, q3, q4, q5, q6]

    def _post_process(self, solutions):
        for q1, q2, q3, q4, q5, q6 in solutions:
            q1 = -1 * q1
            q2 = q2 + math.pi / 2
            q4 = q4 * -1
            q6 = q6 * -1 + math.pi / 2
            yield [q1, q2, q3, q4, q5, q6]


class ABB_IRB4600_40_255Kinematics(SphericalWristKinematics):
    """Analytical IK solver for the ABB IRB4600 40/255 robot."""
    def __init__(self):
        points = [Point(0.175, 0.000, 0.495),
                  Point(0.175, 0.000, 1.590),
                  Point(1.446, 0.000, 1.765),
                  Point(1.581, 0.000, 1.765)]
        super(ABB_IRB4600_40_255Kinematics, self).__init__(points)

    def _pre_process(self, joint_values):
        q1, q2, q3, q4, q5, q6 = joint_values
        q1 = -1 * q1
        q2 = q2 - math.pi / 2
        q3 = q3 + math.pi / 2
        q4 = q4 * -1
        q6 = q6 * -1 + math.pi / 2
        return [q1, q2, q3, q4, q5, q6]

    def _post_process(self, solutions):
        for q1, q2, q3, q4, q5, q6 in solutions:
            q1 = -1 * q1
            q2 = q2 + math.pi / 2
            q3 = q3 - math.pi / 2
            q4 = q4 * -1
            q6 = q6 * -1 + math.pi / 2
            yield [q1, q2, q3, q4, q5, q6]



from __future__ import absolute_import

from .exceptions import CartesianMotionError

from .analytical_inverse_kinematics import AnalyticalInverseKinematics
from .analytical_plan_cartesian_motion import AnalyticalPlanCartesianMotion

from .solvers import (
    OffsetWristKinematics,
    SphericalWristKinematics,
    UR3Kinematics,
    UR3eKinematics,
    UR5Kinematics,
    UR5eKinematics,
    UR10Kinematics,
    UR10eKinematics,
    Staubli_TX260LKinematics,
    ABB_IRB4600_40_255Kinematics,
)
__all__ = [
    # exceptions
    'CartesianMotionError',
    # clients / backend features
    'AnalyticalInverseKinematics',
    'AnalyticalPlanCartesianMotion',
    # solvers
    'OffsetWristKinematics',
    'SphericalWristKinematics',
    'UR3Kinematics',
    'UR3eKinematics',
    'UR5Kinematics',
    'UR5eKinematics',
    'UR10Kinematics',
    'UR10eKinematics',
    'Staubli_TX260LKinematics',
    'ABB_IRB4600_40_255Kinematics',
]



"""
Publishes messages to a ROS topic

COMPAS FAB v0.23.0
"""
import time

from ghpythonlib.componentbase import executingcomponent as component
from roslibpy import Topic
from scriptcontext import sticky as st

from compas_fab.backends.ros import ROSmsg
from compas_fab.ghpython.components import create_id


class ROSTopicPublish(component):
    def RunScript(self, ros_client, topic_name, topic_type, msg):
        if not topic_name:
            raise ValueError('Please specify the name of the topic')
        if not topic_type:
            raise ValueError('Please specify the type of the topic')

        key = create_id(self, 'topic')

        topic = st.get(key, None)

        if ros_client and ros_client.is_connected:
            if not topic:
                topic = Topic(ros_client, topic_name, topic_type)
                topic.advertise()
                time.sleep(0.2)

                st[key] = topic

        self.is_advertised = topic and topic.is_advertised

        if msg:
            msg = ROSmsg.parse(msg, topic_type)
            topic.publish(msg.msg)
            self.Message = 'Message published'
        else:
            if self.is_advertised:
                self.Message = 'Topic advertised'

        return (topic, self.is_advertised)



from .spherical_wrist_kinematics import *  # noqa: F403
from .offset_wrist_kinematics import *  # noqa: F403


PLANNER_BACKENDS = {
    'ur3': UR3Kinematics,  # noqa: F405
    'ur3e': UR3eKinematics,  # noqa: F405
    'ur5': UR5Kinematics,  # noqa: F405
    'ur5e': UR5eKinematics,  # noqa: F405
    'ur10': UR10Kinematics,  # noqa: F405
    'ur10e': UR10eKinematics,  # noqa: F405
    'staubli_tx260l': Staubli_TX260LKinematics,  # noqa: F405
    'abb_irb4600_40_255': ABB_IRB4600_40_255Kinematics,  # noqa: F405
}



"""
Merge two configurations.

COMPAS FAB v0.23.0
"""
from ghpythonlib.componentbase import executingcomponent as component


class ConfigMerge(component):
    def RunScript(self, config_a, config_b):

        if config_a and config_b:
            return config_a.merged(config_b)

        return None



"""
Attach a tool to the robot.

COMPAS FAB v0.23.0
"""
from ghpythonlib.componentbase import executingcomponent as component
from compas_rhino.conversions import RhinoMesh
from compas_rhino.conversions import plane_to_compas_frame

from compas.geometry import Frame
from compas_fab.robots import PlanningScene
from compas_fab.robots import Tool


class AttachToolComponent(component):
    def RunScript(self, robot, visual_mesh, collision_mesh, tcf_plane):
        if robot and robot.client and robot.client.is_connected and visual_mesh:
            if not collision_mesh:
                collision_mesh = visual_mesh

            c_visual_mesh = RhinoMesh.from_geometry(visual_mesh).to_compas()
            c_collision_mesh = RhinoMesh.from_geometry(collision_mesh).to_compas()

            if not tcf_plane:
                frame = Frame.worldXY()
            else:
                frame = plane_to_compas_frame(tcf_plane)
            tool = Tool(c_visual_mesh, frame, c_collision_mesh)

            scene = PlanningScene(robot)
            robot.attach_tool(tool)
            scene.add_attached_tool()

        return robot



"""
Get a zero configuration for a robot.

COMPAS FAB v0.23.0
"""
from ghpythonlib.componentbase import executingcomponent as component


class ConfigZero(component):
    def RunScript(self, robot, group):

        if robot:
            return robot.zero_configuration(group)

        return None



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__all__ = [
    'BackendError',
    'KinematicsError',
    'InverseKinematicsError',
]


class BackendError(Exception):
    """Indicates an exceptional state that caused an error
    within the backend engine."""

    def __init__(self, message):
        super(BackendError, self).__init__(message)


class KinematicsError(BackendError):
    """Indicates a kinematic solver exception."""

    def __init__(self, message):
        super(KinematicsError, self).__init__(message)


class InverseKinematicsError(KinematicsError):
    """Indicates that no IK solution could be found by the kinematic solver."""

    def __init__(self):
        super(InverseKinematicsError, self).__init__("No inverse kinematics solution found.")



"""
********************************************************************************
compas_fab.sensors
********************************************************************************

.. currentmodule:: compas_fab.sensors

Package containing a common interface to integrate sensors of various vendors.

Main classes
------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    SerialSensor

Baumer sensors
--------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    PosCon3D
    PosConCM

Exceptions
----------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ProtocolError
    SensorTimeoutError

"""

from .base import (
    SerialSensor,
)
from .baumer import (
    PosCon3D,
    PosConCM,
)
from .exceptions import (
    ProtocolError,
    SensorTimeoutError
)

__all__ = [
    # base
    'SerialSensor',
    # baumer
    'PosCon3D',
    'PosConCM',
    # exceptions
    'ProtocolError',
    'SensorTimeoutError'
]



from math import sin, cos, fabs, asin, acos, sqrt, atan2, pi
from compas.geometry import Frame
from compas_fab.utilities import sign


def forward_kinematics_offset_wrist(joint_values, params):
    """Forward kinematics function for offset wrist 6-axis robots.

    Parameters
    ----------
    joint_values : list of float
        List of 6 joint values in radians.
    params : list of float
        The offset wrist parameters that specify the robot.

    Returns
    -------
    :class:`compas.geometry.Frame`

    Notes
    -----
    Code adapted from https://github.com/ros-industrial/universal_robot/blob/indigo-devel/ur_kinematics/src/ur_kin.cpp
    """
    d1, a2, a3, d4, d5, d6 = params
    q = joint_values

    s1, c1 = sin(q[0]), cos(q[0])
    q23, q234, s2, c2 = q[1], q[1], sin(q[1]), cos(q[1])
    q23 += q[2]
    q234 += q[2]
    s4, c4 = sin(q[3]), cos(q[3])
    q234 += q[3]
    s5, c5 = sin(q[4]), cos(q[4])
    s6, c6 = sin(q[5]), cos(q[5])
    s23, c23 = sin(q23), cos(q23)
    s234, c234 = sin(q234), cos(q234)

    T = [0. for _ in range(4*4)]

    T[0] = c234*c1*s5 - c5*s1
    T[1] = c6*(s1*s5 + c234*c1*c5) - s234*c1*s6
    T[2] = -s6*(s1*s5 + c234*c1*c5) - s234*c1*c6
    T[3] = d6*c234*c1*s5 - a3*c23*c1 - a2 * \
        c1*c2 - d6*c5*s1 - d5*s234*c1 - d4*s1
    T[4] = c1*c5 + c234*s1*s5
    T[5] = -c6*(c1*s5 - c234*c5*s1) - s234*s1*s6
    T[6] = s6*(c1*s5 - c234*c5*s1) - s234*c6*s1
    T[7] = d6*(c1*c5 + c234*s1*s5) + d4*c1 - \
        a3*c23*s1 - a2*c2*s1 - d5*s234*s1
    T[8] = -s234*s5
    T[9] = -c234*s6 - s234*c5*c6
    T[10] = s234*c5*s6 - c234*c6
    T[11] = d1 + a3*s23 + a2*s2 - d5 * \
        (c23*c4 - s23*s4) - d6*s5*(c23*s4 + s23*c4)
    T[15] = 1.0

    frame = Frame((T[3], T[7], T[11]),
                  (T[0], T[4], T[8]), (T[1], T[5], T[9]))

    return frame


def inverse_kinematics_offset_wrist(frame, params, q6_des=0.):
    """Inverse kinematics function for offset wrist 6-axis robots.

    Parameters
    ----------
    frame : :class:`compas.geometry.Frame`
        The frame we want to calculate the inverse kinematics for.
    params : list of float
        The offset wrist parameters that specify the robot.

    Returns
    -------
    list of list

    Notes
    -----
    Code adapted from https://github.com/ros-industrial/universal_robot/blob/indigo-devel/ur_kinematics/src/ur_kin.cpp
    """

    ZERO_THRESH = 0.00000001
    d1, a2, a3, d4, d5, d6 = params

    solutions = []

    T02, T12, T22 = frame.xaxis
    T00, T10, T20 = frame.yaxis
    T01, T11, T21 = frame.zaxis
    T03, T13, T23 = frame.point
    T02 *= -1
    T03 *= -1
    T12 *= -1
    T13 *= -1
    T20 *= -1
    T21 *= -1

    # shoulder rotate joint (q1)
    # q1[2]
    q1 = [0, 0]
    A = d6 * T12 - T13
    B = d6 * T02 - T03
    R = A * A + B * B
    if(fabs(A) < ZERO_THRESH):
        div = 0.0
        if(fabs(fabs(d4) - fabs(B)) < ZERO_THRESH):
            div = -sign(d4) * sign(B)
        else:
            div = -d4 / B
        arcsin = asin(div)
        if(fabs(arcsin) < ZERO_THRESH):
            arcsin = 0.0
        if(arcsin < 0.0):
            q1[0] = arcsin + 2.0 * pi
        else:
            q1[0] = arcsin
        q1[1] = pi - arcsin

    elif(fabs(B) < ZERO_THRESH):
        div = 0.0
        if(fabs(fabs(d4) - fabs(A)) < ZERO_THRESH):
            div = sign(d4)*sign(A)
        else:
            div = d4 / A
        arccos = acos(div)
        q1[0] = arccos
        q1[1] = 2.0 * pi - arccos

    elif(d4 * d4 > R):
        raise ValueError("No solutions")
    else:
        arccos = acos(d4 / sqrt(R))
        arctan = atan2(-B, A)
        pos = arccos + arctan
        neg = -arccos + arctan
        if(fabs(pos) < ZERO_THRESH):
            pos = 0.0
        if(fabs(neg) < ZERO_THRESH):
            neg = 0.0
        if(pos >= 0.0):
            q1[0] = pos
        else:
            q1[0] = 2.0*pi + pos
        if(neg >= 0.0):
            q1[1] = neg
        else:
            q1[1] = 2.0*pi + neg

    # wrist 2 joint (q5)
    q5 = [[0, 0], [0, 0]]
    for i in range(2):
        numer = (T03 * sin(q1[i]) - T13*cos(q1[i]) - d4)
        div = 0.0
        if(fabs(fabs(numer) - fabs(d6)) < ZERO_THRESH):
            div = sign(numer) * sign(d6)
        else:
            div = numer / d6
        arccos = acos(div)
        q5[i][0] = arccos
        q5[i][1] = 2.0 * pi - arccos

    for i in range(2):
        for j in range(2):
            c1 = cos(q1[i])
            s1 = sin(q1[i])
            c5 = cos(q5[i][j])
            s5 = sin(q5[i][j])
            q6 = 0.0

            # wrist 3 joint (q6)
            if(fabs(s5) < ZERO_THRESH):
                q6 = q6_des
            else:
                q6 = atan2(sign(s5)*-(T01*s1 - T11*c1),
                           sign(s5)*(T00*s1 - T10*c1))
            if(fabs(q6) < ZERO_THRESH):
                q6 = 0.0
            if(q6 < 0.0):
                q6 += 2.0*pi

            # RRR joints (q2,q3,q4)
            q2, q3, q4 = [0, 0], [0, 0], [0, 0]

            c6 = cos(q6)
            s6 = sin(q6)
            x04x = -s5*(T02*c1 + T12*s1) - c5 * \
                (s6*(T01*c1 + T11*s1) - c6*(T00*c1 + T10*s1))
            x04y = c5*(T20*c6 - T21*s6) - T22*s5
            p13x = d5*(s6*(T00*c1 + T10*s1) + c6*(T01*c1 + T11*s1)
                       ) - d6*(T02*c1 + T12*s1) + T03*c1 + T13*s1
            p13y = T23 - d1 - d6*T22 + d5*(T21*c6 + T20*s6)

            c3 = (p13x*p13x + p13y*p13y - a2*a2 - a3*a3) / (2.0*a2*a3)
            if(fabs(fabs(c3) - 1.0) < ZERO_THRESH):
                c3 = sign(c3)
            elif(fabs(c3) > 1.0):
                solutions.extend([None, None])
                continue

            arccos = acos(c3)
            q3[0] = arccos
            q3[1] = 2.0*pi - arccos
            denom = a2*a2 + a3*a3 + 2*a2*a3*c3
            s3 = sin(arccos)
            A = (a2 + a3*c3)
            B = a3*s3
            q2[0] = atan2((A*p13y - B*p13x) / denom,
                          (A*p13x + B*p13y) / denom)
            q2[1] = atan2((A*p13y + B*p13x) / denom,
                          (A*p13x - B*p13y) / denom)
            c23_0 = cos(q2[0]+q3[0])
            s23_0 = sin(q2[0]+q3[0])
            c23_1 = cos(q2[1]+q3[1])
            s23_1 = sin(q2[1]+q3[1])
            q4[0] = atan2(c23_0*x04y - s23_0*x04x, x04x*c23_0 + x04y*s23_0)
            q4[1] = atan2(c23_1*x04y - s23_1*x04x, x04x*c23_1 + x04y*s23_1)

            for k in range(2):
                if(fabs(q2[k]) < ZERO_THRESH):
                    q2[k] = 0.0
                elif(q2[k] < 0.0):
                    q2[k] += 2.0*pi
                if(fabs(q4[k]) < ZERO_THRESH):
                    q4[k] = 0.0
                elif(q4[k] < 0.0):
                    q4[k] += 2.0*pi

                solutions.append([q1[i], q2[k], q3[k], q4[k], q5[i][j], q6])

    return solutions


if __name__ == "__main__":
    from compas.geometry import allclose
    params = [0.089159, -0.42500, -0.39225, 0.10915, 0.09465, 0.0823]  # ur5
    q = [0.2, 5.5, 1.4, 1.3, 2.6, 3.6]
    frame = forward_kinematics_offset_wrist(q, params)
    sol = inverse_kinematics_offset_wrist(frame, params)
    assert(allclose(sol[0], q))



import math
from compas.robots import Configuration


def get_smaller_angle(angle):  # delete
    angle1 = angle % (2 * math.pi)
    angle2 = angle1 - 2 * math.pi
    return angle1 if angle1 < math.fabs(angle2) else angle2


def smallest_joint_angles(joint_angles):  # move to configuration ?
    return [get_smaller_angle(j) for j in joint_angles]


def fit_within_bounds(angle, lower, upper):
    while angle < lower:
        angle += 2 * math.pi
    while angle > upper:
        angle -= 2 * math.pi
    assert(angle >= lower and angle <= upper), "Joint angle out of bounds."
    return angle


def joint_angles_to_configurations(robot, solutions, group=None):
    joint_names = robot.get_configurable_joint_names(group=group)
    return [Configuration.from_revolute_values(q, joint_names=joint_names) if q else None for q in solutions]


def try_to_fit_configurations_between_bounds(robot, configurations, group=None):
    j1, j2, j3, j4, j5, j6 = robot.get_configurable_joints(group=group)
    for i, c in enumerate(configurations):
        if c is None:
            continue
        a1, a2, a3, a4, a5, a6 = c.values()
        try:
            a1 = fit_within_bounds(a1, j1.limit.lower, j1.limit.upper)
            a2 = fit_within_bounds(a2, j2.limit.lower, j2.limit.upper)
            a3 = fit_within_bounds(a3, j3.limit.lower, j3.limit.upper)
            a4 = fit_within_bounds(a4, j4.limit.lower, j4.limit.upper)
            a5 = fit_within_bounds(a5, j5.limit.lower, j5.limit.upper)
            a6 = fit_within_bounds(a6, j6.limit.lower, j6.limit.upper)
            configurations[i].joint_values = [a1, a2, a3, a4, a5, a6]
        except AssertionError:
            configurations[i] = None
    return configurations



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import random

from compas.robots import Joint

from compas_fab.backends.exceptions import InverseKinematicsError
from compas_fab.backends.interfaces import InverseKinematics
from compas_fab.backends.pybullet.conversions import pose_from_frame
from compas_fab.utilities import LazyLoader

pybullet = LazyLoader('pybullet', globals(), 'pybullet')


__all__ = [
    'PyBulletInverseKinematics',
]


class PyBulletInverseKinematics(InverseKinematics):
    """Callable to calculate the robot's inverse kinematics for a given frame."""
    def __init__(self, client):
        self.client = client

    def inverse_kinematics(self, robot, frame_WCF, start_configuration=None, group=None, options=None):
        """Calculate the robot's inverse kinematic for a given frame.

        Parameters
        ----------
        robot : :class:`compas_fab.robots.Robot`
            The robot instance for which inverse kinematics is being calculated.
        frame_WCF: :class:`compas.geometry.Frame`
            The frame to calculate the inverse for.
        start_configuration: :class:`compas_fab.robots.Configuration`, optional
            If passed, the inverse will be calculated such that the calculated
            joint positions differ the least from the start_configuration.
            Defaults to the zero configuration.
        group: str, optional
            The planning group used for determining the end effector and labeling
            the ``start_configuration``. Defaults to the robot's main planning group.
        options: dict, optional
            Dictionary containing the following key-value pairs:

            - ``"link_name"``: (:obj:`str`, optional ) Name of the link for which
              to compute the inverse kinematics.  Defaults to the given group's end
              effector.
            - ``"semi-constrained"``: (:obj:`bool`, optional) When ``True``, only the
              position and not the orientation of ``frame_WCF`` will not be considered
              in the calculation.  Defaults to ``False``.
            - ``"enforce_joint_limits"``: (:obj:`bool`, optional) When ``False``, the
              robot's joint limits will be ignored in the calculation.  Defaults to
              ``True``.
            - ``"high_accuracy"``:  (:obj:`bool`, optional) When ``True``, the
              solver will iteratively try to approach minimum deviation from the requested
              target frame.  Defaults to ``True``.
            - ``"high_accuracy_threshold"``:  (:obj:`float`, optional) Defines the maximum
              acceptable distance threshold for the high accuracy solver. Defaults to ``1e-6``.
            - ``"high_accuracy_max_iter"``:  (:obj:`float`, optional) Defines the maximum
              number of iterations to use for the high accuracy solver. Defaults to ``20``.
            - ``"max_results"``: (:obj:`int`) Maximum number of results to return.
              Defaults to ``100``.

        Yields
        -------
        :obj:`tuple` of :obj:`list`
            A tuple of 2 elements containing a list of joint positions and a list of matching joint names.

        Raises
        ------
        :class:`compas_fab.backends.InverseKinematicsError`
        """
        options = options or {}
        high_accuracy = options.get('high_accuracy', True)
        max_results = options.get('max_results', 100)
        link_name = options.get('link_name') or robot.get_end_effector_link_name(group)
        cached_robot = self.client.get_cached_robot(robot)
        body_id = self.client.get_uid(cached_robot)
        link_id = self.client._get_link_id_by_name(link_name, cached_robot)
        point, orientation = pose_from_frame(frame_WCF)

        joints = cached_robot.get_configurable_joints()
        joints.sort(key=lambda j: j.attr['pybullet']['id'])
        joint_names = [joint.name for joint in joints]

        lower_limits = [joint.limit.lower if joint.type != Joint.CONTINUOUS else 0 for joint in joints]
        upper_limits = [joint.limit.upper if joint.type != Joint.CONTINUOUS else 2 * math.pi for joint in joints]

        start_configuration = start_configuration or robot.zero_configuration(group)
        start_configuration = self.client.set_robot_configuration(robot, start_configuration, group)

        rest_poses = self._get_rest_poses(joint_names, start_configuration)

        for _ in range(max_results):
            ik_options = dict(
                bodyUniqueId=body_id,
                endEffectorLinkIndex=link_id,
                targetPosition=point,
                physicsClientId=self.client.client_id,
            )

            if options.get('enforce_joint_limits', True):
                # I don't know what jointRanges needs to be.  Erwin Coumans knows, but he isn't telling.
                # https://stackoverflow.com/questions/49674179/understanding-inverse-kinematics-pybullet
                # https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/preview?pru=AAABc7276PI*zazLer2rlZ8tAUI8lF98Kw#heading=h.9i02ojf4k3ve
                joint_ranges = [u - l for u, l in zip(upper_limits, lower_limits)]

                if options.get('semi-constrained'):
                    ik_options.update(dict(
                        lowerLimits=lower_limits,
                        upperLimits=upper_limits,
                        jointRanges=joint_ranges,
                        restPoses=rest_poses,
                    ))
                else:
                    ik_options.update(dict(
                        targetPosition=point,
                        targetOrientation=orientation,
                        lowerLimits=lower_limits,
                        upperLimits=upper_limits,
                        jointRanges=joint_ranges,
                        restPoses=rest_poses,
                    ))
            else:
                if not options.get('semi-constrained'):
                    ik_options.update(dict(
                        targetOrientation=orientation,
                    ))

            # Ready to call IK
            if high_accuracy:
                ik_options.update(dict(
                    joints=joints,
                    threshold=options.get('high_accuracy_threshold', 1e-6),
                    max_iter=options.get('high_accuracy_max_iter', 20),
                ))
                joint_positions, close_enough = self._accurate_inverse_kinematics(**ik_options)

                # NOTE: In principle, this accurate iter IK should work out of the
                # pybullet box, but the results seem to be way off target. For now,
                # I'm leaving the legacy iterative accurate ik in python as per
                # older examples of pybullet, until we figure out why the builtin
                # one is not cooperating.

                # ik_options.update(dict(
                #     residualThreshold=options.get('high_accuracy_threshold', 1e-6),
                #     maxNumIterations=options.get('high_accuracy_max_iter', 20),
                # ))
                # joint_positions = pybullet.calculateInverseKinematics(**ik_options)
            else:
                joint_positions = pybullet.calculateInverseKinematics(**ik_options)
                close_enough = True     # yeah, let's say we trust it

            if not joint_positions:
                raise InverseKinematicsError()

            # Randomize joints to get a different solution on the next iter
            self.client._set_joint_positions([joint.attr['pybullet']['id'] for joint in joints],
                                             [random.uniform(*limits) for limits in zip(lower_limits, upper_limits)],
                                             body_id)

            # Normally, no results ends the iteration, but if we have no solution
            # because the accurate IK solving says it's not close enough, we can retry
            # with a new randomized seed
            if not close_enough and high_accuracy:
                continue

            rest_poses = joint_positions
            yield joint_positions, joint_names

    def _accurate_inverse_kinematics(self, joints, threshold, max_iter, **kwargs):
        # Based on these examples
        # https://github.com/bulletphysics/bullet3/blob/master/examples/pybullet/examples/inverse_kinematics_husky_kuka.py#L81
        close_enough = False
        iter = 0
        distance = None
        joint_ids = [joint.attr['pybullet']['id'] for joint in joints]
        body_id = kwargs['bodyUniqueId']
        link_id = kwargs['endEffectorLinkIndex']
        target_position = kwargs['targetPosition']

        while not close_enough and iter < max_iter:
            joint_poses = pybullet.calculateInverseKinematics(**kwargs)
            for i in range(len(joint_ids)):
                pybullet.resetJointState(body_id, joint_ids[i], joint_poses[i])

            link_state = pybullet.getLinkState(body_id, link_id)
            new_pose = link_state[4]

            diff = [target_position[0] - new_pose[0], target_position[1] - new_pose[1], target_position[2] - new_pose[2]]
            distance = (diff[0] * diff[0] + diff[1] * diff[1] + diff[2] * diff[2])
            close_enough = distance < threshold
            kwargs['restPoses'] = joint_poses
            iter += 1

        return joint_poses, close_enough

    def _get_rest_poses(self, joint_names, configuration):
        name_value_map = {configuration.joint_names[i]: configuration.joint_values[i] for i in range(len(configuration.joint_names))}
        return [name_value_map[name] for name in joint_names]
